from typing import List

from ci_pipe.step import Step


class Branch:
    def __init__(self, name, steps):
        self._name = name
        self._steps = steps

    @classmethod
    def from_dict(cls, name, data):
        serialized_steps = data.get("steps", [])
        steps = [Step.from_dict(serialized_step) for serialized_step in serialized_steps]
        return cls(name, steps)

    def to_dict(self):
        return {
            "steps": [
                {
                    "index": index,
                    "name": step.name(),
                    "params": step.arguments(),
                    "outputs": step.step_output(),
                }
                for index, step in enumerate(self._steps, start=1)
            ]
        }

    def add_steps(self, steps: List[Step]):
        self._steps.extend(steps)

    def name(self):
        return self._name

    def steps(self):
        return self._steps



from ci_pipe.decorators import step
from ci_pipe.errors.caiman_backend_not_configured_error import CaimanBackendNotConfiguredError


class CaimanModule:
    MOTION_CORRECTION_STEP = "Caiman Motion Correction"
    MOTION_CORRECTION_VIDEOS_SUFFIX = "MC"
    CNMF_STEP = "Caiman Constrained Non-negative Matrix Factorization"
    CNMF_VIDEOS_SUFFIX = "CNMF"

    def __init__(self, caiman, ci_pipe):
        if caiman is None:
            raise CaimanBackendNotConfiguredError()
        self._caiman = caiman
        self._ci_pipe = ci_pipe

    @step(MOTION_CORRECTION_STEP)
    def motion_correction(
            self,
            inputs,
            *,
            caiman_strides=(48, 48),
            caiman_overlaps=(24, 24),
            caiman_max_shifts=(6, 6),
            caiman_max_deviation_rigid=3,
            caiman_pw_rigid=True,
            caiman_shifts_opencv=True,
            caiman_border_nan='copy',
            caiman_save_movie=True
    ):
        # TODO: Think if we should grab all potential extensions accepted by motion correction
        output = []
        output_dir = self._ci_pipe.create_output_directory_for_next_step(self.MOTION_CORRECTION_STEP)

        for input_data in inputs('videos-tiff'):
            motion_correct_handler = self._caiman.motion_correction.MotionCorrect(
                fname=input_data['value'],
                strides=caiman_strides,
                overlaps=caiman_overlaps,
                max_shifts=caiman_max_shifts,
                max_deviation_rigid=caiman_max_deviation_rigid,
                pw_rigid=caiman_pw_rigid,
                shifts_opencv=caiman_shifts_opencv,
                border_nan=caiman_border_nan,
            )
            motion_correct_handler.motion_correct(save_movie=caiman_save_movie)
            mmap_files = motion_correct_handler.mmap_file
            mmap_path = mmap_files[0] # we are processing only one at a time, that's why we can unpack it like this

            memmapped_movie = self._caiman.load(mmap_path)

            tif_output_path = self._ci_pipe.make_output_file_path(
                mmap_path,
                output_dir,
                self.MOTION_CORRECTION_VIDEOS_SUFFIX,
                ext="tif",
            )

            memmapped_movie.save(tif_output_path)
            output.append({'ids': input_data['ids'], 'value': tif_output_path})

        return {"videos-tiff": output}

    @step(CNMF_STEP)
    def cnmf(
            self,
            inputs,
            *,
            caiman_n_processes=None,
            caiman_k=5,
            caiman_gSig=[4, 4],
            caiman_gSiz=None,
            caiman_merge_thresh=0.8,
            caiman_p=2,
            caiman_dview=None,
            caiman_Ain=None,
            caiman_Cin=None,
            caiman_b_in=None,
            caiman_f_in=None,
            caiman_do_merge=True,
            caiman_ssub=2,
            caiman_tsub=2,
            caiman_p_ssub=1,
            caiman_p_tsub=1,
            caiman_method_init='greedy_roi',
            caiman_alpha_snmf=0.5,
            caiman_rf=None,
            caiman_stride=None,
            caiman_memory_fact=1,
            caiman_gnb=1,
            caiman_nb_patch=1,
            caiman_only_init_patch=False,
            caiman_method_deconvolution='oasis',
            caiman_n_pixels_per_process=4000,
            caiman_block_size_temp=5000,
            caiman_num_blocks_per_run_temp=20,
            caiman_num_blocks_per_run_spat=20,
            caiman_check_nan=True,
            caiman_skip_refinement=False,
            caiman_normalize_init=True,
            caiman_options_local_NMF=None,
            caiman_minibatch_shape=100,
            caiman_minibatch_suff_stat=3,
            caiman_update_num_comps=True,
            caiman_rval_thr=0.9,
            caiman_thresh_fitness_delta=-20,
            caiman_thresh_fitness_raw=None,
            caiman_thresh_overlap=.5,
            caiman_batch_update_suff_stat=False,
            caiman_s_min=None,
            caiman_remove_very_bad_comps=False,
            caiman_border_pix=0,
            caiman_low_rank_background=True,
            caiman_update_background_components=True,
            caiman_rolling_sum=True,
            caiman_rolling_length=100,
            caiman_min_corr=.85,
            caiman_min_pnr=20,
            caiman_ring_size_factor=1.5,
            caiman_center_psf=False,
            caiman_use_dense=True,
            caiman_deconv_flag=True,
            caiman_simultaneously=False,
            caiman_n_refit=0,
            caiman_del_duplicates=False,
            caiman_N_samples_exceptionality=None,
            caiman_max_num_added=3,
            caiman_min_num_trial=2,
            caiman_thresh_CNN_noisy=0.5,
            caiman_fr=30,
            caiman_decay_time=0.4,
            caiman_min_SNR=2.5,
            caiman_ssub_B=2,
            caiman_init_iter=2,
            caiman_sniper_mode=False,
            caiman_use_peak_max=False,
            caiman_test_both=False,
            caiman_expected_comps=500,
            caiman_params=None
    ):
        output = []
        output_dir = self._ci_pipe.create_output_directory_for_next_step(self.CNMF_STEP)

        for input_data in inputs('videos-tiff'):
            cnmf_model = self._caiman.source_extraction.cnmf.CNMF(
                n_processes=caiman_n_processes,
                k=caiman_k,
                gSig=caiman_gSig,
                gSiz=caiman_gSiz,
                merge_thresh=caiman_merge_thresh,
                p=caiman_p,
                dview=caiman_dview,
                Ain=caiman_Ain,
                Cin=caiman_Cin,
                b_in=caiman_b_in,
                f_in=caiman_f_in,
                do_merge=caiman_do_merge,
                ssub=caiman_ssub,
                tsub=caiman_tsub,
                p_ssub=caiman_p_ssub,
                p_tsub=caiman_p_tsub,
                method_init=caiman_method_init,
                alpha_snmf=caiman_alpha_snmf,
                rf=caiman_rf,
                stride=caiman_stride,
                memory_fact=caiman_memory_fact,
                gnb=caiman_gnb,
                nb_patch=caiman_nb_patch,
                only_init_patch=caiman_only_init_patch,
                method_deconvolution=caiman_method_deconvolution,
                n_pixels_per_process=caiman_n_pixels_per_process,
                block_size_temp=caiman_block_size_temp,
                num_blocks_per_run_temp=caiman_num_blocks_per_run_temp,
                num_blocks_per_run_spat=caiman_num_blocks_per_run_spat,
                check_nan=caiman_check_nan,
                skip_refinement=caiman_skip_refinement,
                normalize_init=caiman_normalize_init,
                options_local_NMF=caiman_options_local_NMF,
                minibatch_shape=caiman_minibatch_shape,
                minibatch_suff_stat=caiman_minibatch_suff_stat,
                update_num_comps=caiman_update_num_comps,
                rval_thr=caiman_rval_thr,
                thresh_fitness_delta=caiman_thresh_fitness_delta,
                thresh_fitness_raw=caiman_thresh_fitness_raw,
                thresh_overlap=caiman_thresh_overlap,
                batch_update_suff_stat=caiman_batch_update_suff_stat,
                s_min=caiman_s_min,
                remove_very_bad_comps=caiman_remove_very_bad_comps,
                border_pix=caiman_border_pix,
                low_rank_background=caiman_low_rank_background,
                update_background_components=caiman_update_background_components,
                rolling_sum=caiman_rolling_sum,
                rolling_length=caiman_rolling_length,
                min_corr=caiman_min_corr,
                min_pnr=caiman_min_pnr,
                ring_size_factor=caiman_ring_size_factor,
                center_psf=caiman_center_psf,
                use_dense=caiman_use_dense,
                deconv_flag=caiman_deconv_flag,
                simultaneously=caiman_simultaneously,
                n_refit=caiman_n_refit,
                del_duplicates=caiman_del_duplicates,
                N_samples_exceptionality=caiman_N_samples_exceptionality,
                max_num_added=caiman_max_num_added,
                min_num_trial=caiman_min_num_trial,
                thresh_CNN_noisy=caiman_thresh_CNN_noisy,
                fr=caiman_fr,
                decay_time=caiman_decay_time,
                min_SNR=caiman_min_SNR,
                ssub_B=caiman_ssub_B,
                init_iter=caiman_init_iter,
                sniper_mode=caiman_sniper_mode,
                use_peak_max=caiman_use_peak_max,
                test_both=caiman_test_both,
                expected_comps=caiman_expected_comps,
                params=caiman_params
            )

            # Note: Values for this algorithm are changed within estimates object of cnmf model
            memmapped_movie = self._caiman.load(input_data['value'])
            cnmf_model.fit(images=memmapped_movie)

            hdf5_output_path = self._ci_pipe.make_output_file_path(
                input_data['value'],
                output_dir,
                self.CNMF_VIDEOS_SUFFIX,
                ext="hdf5",
            )

            cnmf_model.save(hdf5_output_path)
            output.append({'ids': input_data['ids'], 'value': hdf5_output_path})

        return {"files-hdf5": output}



class CIPipeError(Exception):
    def __init__(self, message, context=None):
        self._message = message
        self.context = context or {}
        super().__init__(self._message)

    def to_dict(self):
        return {"message": str(self._message), "context": self.context}


from pathlib import Path

import numpy as np

from ci_pipe.decorators import step
from ci_pipe.errors.isx_backend_not_configured_error import ISXBackendNotConfiguredError
from ci_pipe.utils.project_template import load_project_templates


class ISXModule:
    # TODO: Consider moving these constants to a different config file
    PREPROCESS_VIDEOS_STEP = "ISX Preprocess Videos"
    BANDPASS_FILTER_VIDEOS_STEP = "ISX Bandpass Filter Videos"
    MOTION_CORRECTION_VIDEOS_STEP = "ISX Motion Correction Videos"
    NORMALIZE_DFF_VIDEOS_STEP = "ISX Normalize DFF Videos"
    EXTRACT_NEURONS_PCA_ICA_STEP = "ISX Extract Neurons PCA ICA"
    DETECT_EVENTS_IN_CELLS_STEP = "ISX Detect Events In Cells"
    AUTO_ACCEPT_REJECT_CELLS_STEP = "ISX Auto Accept Reject Cells"
    EXPORT_MOVIE_TO_TIFF_STEP = "ISX Export movie to TIFF"
    EXPORT_MOVIE_TO_NWB_STEP = "ISX Export movie to NWB"
    LONGITUDINAL_REGISTRATION_STEP = "ISX Longitudinal Registration"
    PREPROCESS_VIDEOS_SUFFIX = "PP"
    BANDPASS_FILTER_VIDEOS_SUFFIX = "BP"
    MOTION_CORRECTION_VIDEOS_SUFFIX = "MC"
    MOTION_CORRECTION_VIDEOS_TRANSLATIONS_SUFFIX = "translations"
    MOTION_CORRECTION_VIDEOS_CROP_RECT_SUFFIX = "crop-rect"
    MOTION_CORRECTION_VIDEOS_MEAN_IMAGES_SUFFIX = "mean-image"
    NORMALIZE_DFF_VIDEOS_SUFFIX = "DFF"
    EXTRACT_NEURONS_PCA_ICA_VIDEOS_SUFFIX = "PCA-ICA"
    DETECT_EVENTS_IN_CELLS_SUFFIX = "ED"
    LONGITUDINAL_REGISTRATION_SUFFIX = "LR"
    LONGITUDINAL_REGISTRATION_CORRESPONDENCES_TABLE_NAME = "LR-correspondences-table"
    LONGITUDINAL_REGISTRATION_CROP_RECT_NAME = "LR-crop-rect"
    LONGITUDINAL_REGISTRATION_TRANSFORM_NAME = "LR-transform"
    GUI_VISUALIZATION_STEP = "ISX Gui Visualization"
    GUI_VISUALIZATION_SUFFIX = "GUI"

    def __init__(self, isx, ci_pipe):
        if isx is None:
            raise ISXBackendNotConfiguredError()
        self._isx = isx
        self._ci_pipe = ci_pipe

        self._lr_reference_selection_strategies = {
            'by_num_cells_desc': self._lr_by_num_cells_desc,
        }

    # TODO: Find the best way to remove repetition on these step methods, without losing clarity of what each step does

    @step(PREPROCESS_VIDEOS_STEP)
    def preprocess_videos(
            self,
            inputs,
            *,
            isx_pp_temporal_downsample_factor=1,
            isx_pp_spatial_downsample_factor=1,
            isx_pp_crop_rect=None,
            isx_pp_crop_rect_format="tlbr",
            isx_pp_fix_defective_pixels=True,
            isx_pp_trim_early_frames=True
    ):
        output = []
        output_dir = self._ci_pipe.create_output_directory_for_next_step(self.PREPROCESS_VIDEOS_STEP)

        for input in inputs('videos-isxd'):
            input_path = input['value']
            output_path = self._isx.make_output_file_path(input_path, output_dir, self.PREPROCESS_VIDEOS_SUFFIX)

            self._isx.preprocess(
                input_movie_files=[input_path],
                output_movie_files=[output_path],
                temporal_downsample_factor=isx_pp_temporal_downsample_factor,
                spatial_downsample_factor=isx_pp_spatial_downsample_factor,
                crop_rect=isx_pp_crop_rect,
                crop_rect_format=isx_pp_crop_rect_format,
                fix_defective_pixels=isx_pp_fix_defective_pixels,
                trim_early_frames=isx_pp_trim_early_frames
            )

            output.append({'ids': input['ids'], 'value': output_path})

        return {
            'videos-isxd': output
        }

    @step(BANDPASS_FILTER_VIDEOS_STEP)
    def bandpass_filter_videos(
            self,
            inputs,
            *,
            isx_bp_low_cutoff=0.005,
            isx_bp_high_cutoff=0.5,
            isx_bp_retain_mean=False,
            isx_bp_subtract_global_minimum=True
    ):
        output = []
        output_dir = self._ci_pipe.create_output_directory_for_next_step(self.BANDPASS_FILTER_VIDEOS_STEP)

        for input in inputs('videos-isxd'):
            input_path = input['value']
            output_path = self._isx.make_output_file_path(input_path, output_dir, self.BANDPASS_FILTER_VIDEOS_SUFFIX)

            self._isx.spatial_filter(
                input_movie_files=[input_path],
                output_movie_files=[output_path],
                low_cutoff=isx_bp_low_cutoff,
                high_cutoff=isx_bp_high_cutoff,
                retain_mean=isx_bp_retain_mean,
                subtract_global_minimum=isx_bp_subtract_global_minimum
            )

            output.append({'ids': input['ids'], 'value': output_path})

        return {
            'videos-isxd': output
        }

    @step(MOTION_CORRECTION_VIDEOS_STEP)
    def motion_correction_videos(
            self,
            inputs,
            *,
            isx_mc_series_name="series",
            isx_mc_max_translation=20,
            isx_mc_low_bandpass_cutoff=0.004,
            isx_mc_high_bandpass_cutoff=0.016,
            isx_mc_roi=None,
            isx_mc_reference_segment_index=0,
            isx_mc_reference_frame_index=0,
            isx_mc_global_registration_weight=1,
            isx_mc_preserve_input_dimensions=False
    ):
        output_videos = []
        output_translations = []
        output_crop_rects = []
        output_mean_images = []
        output_dir = self._ci_pipe.create_output_directory_for_next_step(self.MOTION_CORRECTION_VIDEOS_STEP)

        for input in inputs('videos-isxd'):
            input_path = input['value']
            output_video_path = self._isx.make_output_file_path(input_path, output_dir,
                                                                self.MOTION_CORRECTION_VIDEOS_SUFFIX)
            output_translations_path = self._isx.make_output_file_path(input_path, output_dir,
                                                                       self.MOTION_CORRECTION_VIDEOS_TRANSLATIONS_SUFFIX,
                                                                       ext='csv')
            output_crop_rect_path = self._isx.make_output_file_path(input_path, output_dir,
                                                                    f'{isx_mc_series_name}-{self.MOTION_CORRECTION_VIDEOS_CROP_RECT_SUFFIX}',
                                                                    ext='csv')
            output_mean_image_path = self._isx.make_output_file_path(input_path, output_dir,
                                                                     f'{isx_mc_series_name}-{self.MOTION_CORRECTION_VIDEOS_MEAN_IMAGES_SUFFIX}')

            self._isx.project_movie(
                input_movie_files=[input_path],
                output_image_file=output_mean_image_path
            )

            self._isx.motion_correct(
                input_movie_files=[input_path],
                output_movie_files=[output_video_path],
                max_translation=isx_mc_max_translation,
                low_bandpass_cutoff=isx_mc_low_bandpass_cutoff,
                high_bandpass_cutoff=isx_mc_high_bandpass_cutoff,
                roi=isx_mc_roi,
                reference_segment_index=isx_mc_reference_segment_index,
                reference_frame_index=isx_mc_reference_frame_index,
                reference_file_name=output_mean_image_path,
                global_registration_weight=isx_mc_global_registration_weight,
                output_translation_files=[output_translations_path],
                output_crop_rect_file=output_crop_rect_path,
                preserve_input_dimensions=isx_mc_preserve_input_dimensions
            )

            output_videos.append({'ids': input['ids'], 'value': output_video_path})
            output_translations.append({'ids': input['ids'], 'value': output_translations_path})
            output_crop_rects.append({'ids': input['ids'], 'value': output_crop_rect_path})
            output_mean_images.append({'ids': input['ids'], 'value': output_mean_image_path})

        return {
            'videos-isxd': output_videos,
            'motion-correction-translations': output_translations,
            'motion-correction-crop-rect': output_crop_rects,
            'motion-correction-mean-images': output_mean_images
        }

    @step(NORMALIZE_DFF_VIDEOS_STEP)
    def normalize_dff_videos(
            self,
            inputs,
            *,
            isx_dff_f0_type='mean'
    ):
        output = []
        output_dir = self._ci_pipe.create_output_directory_for_next_step(self.NORMALIZE_DFF_VIDEOS_STEP)

        for input in inputs('videos-isxd'):
            input_path = input['value']
            output_path = self._isx.make_output_file_path(input_path, output_dir, self.NORMALIZE_DFF_VIDEOS_SUFFIX)

            self._isx.dff(
                input_movie_files=[input_path],
                output_movie_files=[output_path],
                f0_type=isx_dff_f0_type
            )

            output.append({'ids': input['ids'], 'value': output_path})

        return {
            'videos-isxd': output
        }

    @step(EXTRACT_NEURONS_PCA_ICA_STEP)
    def extract_neurons_pca_ica(
            self,
            inputs,
            *,
            isx_pca_ica_num_pcs=180,
            isx_pca_ica_num_ics=120,
            isx_pca_ica_unmix_type='spatial',
            isx_pca_ica_max_iterations=100,
            isx_pca_ica_convergence_threshold=0.00001,
            isx_pca_ica_block_size=1000,
            isx_pca_ica_auto_estimate_num_ics=False,
            isx_pca_ica_average_cell_diameter=13,
    ):
        output = []
        output_dir = self._ci_pipe.create_output_directory_for_next_step(self.EXTRACT_NEURONS_PCA_ICA_STEP)

        for input in inputs('videos-isxd'):
            input_path = input['value']
            output_path = self._isx.make_output_file_path(input_path, output_dir,
                                                          self.EXTRACT_NEURONS_PCA_ICA_VIDEOS_SUFFIX)

            self._isx.pca_ica(
                input_movie_files=[input_path],
                output_cell_set_files=[output_path],
                num_pcs=isx_pca_ica_num_pcs,
                num_ics=isx_pca_ica_num_ics,
                unmix_type=isx_pca_ica_unmix_type,
                max_iterations=isx_pca_ica_max_iterations,
                convergence_threshold=isx_pca_ica_convergence_threshold,
                block_size=isx_pca_ica_block_size,
                auto_estimate_num_ics=isx_pca_ica_auto_estimate_num_ics,
                average_cell_diameter=isx_pca_ica_average_cell_diameter
            )

            output.append({'ids': input['ids'], 'value': output_path})

        return {
            'cellsets-isxd': output
        }

    @step(DETECT_EVENTS_IN_CELLS_STEP)
    def detect_events_in_cells(
            self,
            inputs,
            *,
            isx_ed_threshold=5,
            isx_ed_tau=0.2,
            isx_ed_event_time_ref='beginning',
            isx_ed_ignore_negative_transients=True,
            isx_ed_accepted_cells_only=False
    ):
        output = []
        output_dir = self._ci_pipe.create_output_directory_for_next_step(self.DETECT_EVENTS_IN_CELLS_STEP)

        for input in inputs('cellsets-isxd'):
            input_path = input['value']
            output_path = self._isx.make_output_file_path(input_path, output_dir, self.DETECT_EVENTS_IN_CELLS_SUFFIX)

            self._isx.event_detection(
                input_cell_set_files=[input_path],
                output_event_set_files=[output_path],
                threshold=isx_ed_threshold,
                tau=isx_ed_tau,
                event_time_ref=isx_ed_event_time_ref,
                ignore_negative_transients=isx_ed_ignore_negative_transients,
                accepted_cells_only=isx_ed_accepted_cells_only
            )

            output.append({'ids': input['ids'], 'value': output_path})

        return {
            'events-isxd': output
        }

    @step(AUTO_ACCEPT_REJECT_CELLS_STEP)
    def auto_accept_reject_cells(
            self,
            inputs,
            *,
            isx_acr_filters=None
    ):
        output = []
        self._ci_pipe.create_output_directory_for_next_step(self.AUTO_ACCEPT_REJECT_CELLS_STEP)

        pairs = self._ci_pipe.associate_keys_by_id('cellsets-isxd', 'events-isxd')

        for ids, cellset_path, event_path in pairs:
            output_path = self._ci_pipe.copy_file_to_output_directory(cellset_path, self.AUTO_ACCEPT_REJECT_CELLS_STEP)

            self._isx.auto_accept_reject(
                input_cell_set_files=[output_path],
                input_event_set_files=[event_path],
                filters=isx_acr_filters
            )
            output.append({'ids': ids, 'value': output_path})

        return {
            'cellsets-isxd': output
        }

    @step(EXPORT_MOVIE_TO_TIFF_STEP)
    def export_movie_to_tiff(
            self,
            inputs,
            *,
            isx_emt_write_invalid_frames=False
    ):

        output = []
        output_dir = self._ci_pipe.create_output_directory_for_next_step(self.EXPORT_MOVIE_TO_TIFF_STEP)

        for video in inputs('videos-isxd'):
            input_path = video['value']
            output_path = self._isx.make_output_file_path(input_path, output_dir, '', ext='tiff')

            self._isx.export_movie_to_tiff([input_path], output_path, write_invalid_frames=isx_emt_write_invalid_frames)

            output.append({'ids': video['ids'], 'value': output_path})

        return {
            'videos-tiff': output
        }

    @step(EXPORT_MOVIE_TO_NWB_STEP)
    def export_movie_to_nwb(
            self,
            inputs,
            *,
            isx_emn_write_invalid_frames=False
    ):

        output = []
        output_dir = self._ci_pipe.create_output_directory_for_next_step(self.EXPORT_MOVIE_TO_NWB_STEP)

        for video in inputs('videos-isxd'):
            input_path = video['value']
            output_path = self._isx.make_output_file_path(input_path, output_dir, '', ext='nwb')

            self._isx.export_movie_to_nwb([input_path], output_path, write_invalid_frames=isx_emn_write_invalid_frames)

            output.append({'ids': video['ids'], 'value': output_path})

        return {
            'videos-nwb': output
        }

    @step(LONGITUDINAL_REGISTRATION_STEP)
    def longitudinal_registration(
            self,
            inputs,
            *,
            isx_lr_reference_selection_strategy=None,
            isx_lr_min_correlation=0.5,
            isx_lr_accepted_cells_only=False
    ):
        output_videos = []
        output_cellsets = []
        input_video_paths = []
        output_video_paths = []
        input_cellset_paths = []
        output_cellset_paths = []

        output_dir = self._ci_pipe.create_output_directory_for_next_step(self.LONGITUDINAL_REGISTRATION_STEP)
        all_ids = list(set(id for input in inputs('videos-isxd') for id in input['ids']))

        self._load_outputs_and_paths_from_inputs(inputs('videos-isxd'), output_dir, 'isxd', output_videos,
                                                 input_video_paths, output_video_paths)
        self._load_outputs_and_paths_from_inputs(inputs('cellsets-isxd'), output_dir, 'isxd', output_cellsets,
                                                 input_cellset_paths, output_cellset_paths)

        # TODO: Throw error if strategy name is invalid, consider using enum
        if isx_lr_reference_selection_strategy is not None and isx_lr_reference_selection_strategy in self._lr_reference_selection_strategies:
            input_cellset_paths, output_cellset_paths, input_video_paths, output_video_paths = self._apply_lr_reference_selection(
                isx_lr_reference_selection_strategy,
                input_cellset_paths,
                output_cellset_paths,
                input_video_paths,
                output_video_paths
            )

        output_correspondences_table_path, output_crop_rect_path, output_transform_path = self._generate_lr_output_file_paths()

        self._isx.longitudinal_registration(
            input_cell_set_files=input_cellset_paths,
            output_cell_set_files=output_cellset_paths,
            input_movie_files=input_video_paths,
            output_movie_files=output_video_paths,
            csv_file=output_correspondences_table_path,
            min_correlation=isx_lr_min_correlation,
            accepted_cells_only=isx_lr_accepted_cells_only,
            transform_csv_file=output_transform_path,
            crop_csv_file=output_crop_rect_path
        )

        return {
            'videos-isxd': output_videos,
            'cellsets-isxd': output_cellsets,
            'longitudinal-registration-correspondences-table': [
                {'ids': all_ids, 'value': output_correspondences_table_path}],
            'longitudinal-registration-crop-rect': [{'ids': all_ids, 'value': output_crop_rect_path}],
            'longitudinal-registration-transform': [{'ids': all_ids, 'value': output_transform_path}]
        }

    # Special Gui Visualizer function
    @step(GUI_VISUALIZATION_STEP)
    def create_inscopix_project(self, inputs, *, isx_cellsetname="pca-ica"):
        file_system = self._ci_pipe.file_system()
        output_dir = self._ci_pipe.create_output_directory_for_next_step(self.GUI_VISUALIZATION_STEP)

        dff_by_ids = self._group_by_ids(inputs("videos-isxd"))
        cell_by_ids = self._group_by_ids(inputs("cellsets-isxd"))
        ev_by_ids = self._group_by_ids(inputs("events-isxd"))

        common_unique_ids = sorted(set(dff_by_ids) & set(cell_by_ids) & set(ev_by_ids))
        outputs = []

        plane_template, project_template = load_project_templates()

        for ids_key in common_unique_ids:
            # sort to create stable "plane order"
            dffs = sorted(dff_by_ids[ids_key])
            cellsets = sorted(cell_by_ids[ids_key])
            events = sorted(ev_by_ids[ids_key])

            # pick base name from first dff file
            base_path = Path(dffs[0]).name
            original_movie_name = base_path.replace(f"-{self.NORMALIZE_DFF_VIDEOS_SUFFIX}.isxd", "")
            project_file = file_system.join(output_dir, f"{original_movie_name}-{self.GUI_VISUALIZATION_SUFFIX}.isxp")

            data_folder = project_file[:-5] + "_data"
            file_system.makedirs(data_folder, exist_ok=True)

            planes = []
            for dff, cellset, event in zip(dffs, cellsets, events):
                dmin, dmax = self._movie_first_frame_min_max(dff)

                parsed = plane_template
                replacements = {
                    "{eventdet_path}": event.replace("\\", "/"),
                    "{eventdet_name}": Path(event).name,
                    "{cellset_path}": cellset.replace("\\", "/"),
                    "{cellset_name}": Path(cellset).name,
                    "{DFF_path}": dff.replace("\\", "/"),
                    "{DFF_name}": Path(dff).name,
                    "{dmax}": str(dmax),
                    "{dmin}": str(dmin),
                }
                for k, v in replacements.items():
                    parsed = parsed.replace(k, v)

                planes.append(parsed)

            project_text = project_template.replace(
                "{prj_name}", Path(project_file).name).replace(
                "{plane_1ist}", ", ".join(planes)
            )

            file_system.write(project_file, project_text)
            outputs.append({"ids": list(ids_key), "value": project_file})

        return {"inscopix-projects": outputs}

    # LR reference selection

    def _lr_by_num_cells_desc(self, input_cellsets):
        cell_counts = [self._isx.CellSet.read(path).num_cells for path in input_cellsets]
        return sorted(range(len(cell_counts)), key=lambda i: cell_counts[i], reverse=True)

    # Private methods

    def _load_outputs_and_paths_from_inputs(self, inputs, output_dir, ext, outputs, input_paths, output_paths):
        for input in inputs:
            input_paths.append(input['value'])
            output_path = self._isx.make_output_file_path(input['value'], output_dir,
                                                          self.LONGITUDINAL_REGISTRATION_SUFFIX, ext=ext)
            outputs.append({'ids': input['ids'], 'value': output_path})
            output_paths.append(output_path)

    def _generate_lr_output_file_paths(self):
        output_correspondences_table_path = self._ci_pipe.file_in_output_directory(
            f"{self.LONGITUDINAL_REGISTRATION_CORRESPONDENCES_TABLE_NAME}.csv",
            self.LONGITUDINAL_REGISTRATION_STEP
        )
        output_crop_rect_path = self._ci_pipe.file_in_output_directory(
            f"{self.LONGITUDINAL_REGISTRATION_CROP_RECT_NAME}.csv",
            self.LONGITUDINAL_REGISTRATION_STEP
        )
        output_transform_path = self._ci_pipe.file_in_output_directory(
            f"{self.LONGITUDINAL_REGISTRATION_TRANSFORM_NAME}.csv",
            self.LONGITUDINAL_REGISTRATION_STEP
        )

        return output_correspondences_table_path, output_crop_rect_path, output_transform_path

    def _apply_lr_reference_selection(self, select_reference_lambda, input_cellsets, output_cellsets, input_movies,
                                      output_movies):
        result = self._lr_reference_selection_strategies[select_reference_lambda](input_cellsets)

        if all(isinstance(_, int) for _ in result):
            order = result
        else:
            order = [input_cellsets.index(p) for p in result]

        reorder = lambda lst: [lst[i] for i in order]

        return (
            reorder(input_cellsets),
            reorder(output_cellsets),
            reorder(input_movies),
            reorder(output_movies),
        )

    def _group_by_ids(self, items):
        groups = {}
        for item in items:
            groups.setdefault(tuple(item["ids"]), []).append(item["value"])
        return groups

    def _movie_first_frame_min_max(self, movie_path):
        movie = None
        try:
            movie = self._isx.Movie.read(str(movie_path))
            frame0 = movie.get_frame_data(0)
            if frame0 is None:
                raise ValueError(f"Could not read first frame from: {movie_path}")

            arr = np.asarray(frame0)
            if arr.size == 0:
                raise ValueError(f"First frame is empty for: {movie_path}")

            dmin = float(np.nanmin(arr))
            dmax = float(np.nanmax(arr))

            if not np.isfinite(dmin) or not np.isfinite(dmax):
                raise ValueError(
                    f"Non-finite min/max from first frame for: {movie_path} (dmin={dmin}, dmax={dmax})"
                )

            return dmin, dmax

        except ValueError:
            raise
        except Exception as e:
            raise RuntimeError(f"Failed to compute first-frame min/max for {movie_path}: {e}") from e
        finally:
            try:
                del movie
            except Exception:
                pass



from collections import defaultdict


class SchemaValidator:
    ROOT_PATH = "$"

    @classmethod
    def new_for(cls, schema):
        return cls(schema)

    def __init__(self, schema):
        self._schema = schema

    def validate(self, json_data):
        schema_keys_by_parent = self._collect_child_keys_by_parent(self._schema.as_dict())
        json_keys_by_parent = self._collect_child_keys_by_parent(json_data)

        for parent_path, schema_child_keys in schema_keys_by_parent.items():
            json_child_keys = json_keys_by_parent.get(parent_path, set())
            missing_keys = schema_child_keys - json_child_keys
            if missing_keys:
                raise ValueError(f"{parent_path}: missing keys {missing_keys}")

        schema_values_by_path = self._collect_values_by_path(self._schema.as_dict())
        json_values_by_path = self._collect_values_by_path(json_data)

        for path, schema_values in schema_values_by_path.items():
            json_values = json_values_by_path.get(path, set())
            missing_values = schema_values - json_values
            if missing_values:
                pretty = ", ".join(repr(v) for v in sorted(missing_values, key=lambda x: (type(x).__name__, str(x))))
                raise ValueError(f"{path}: missing values {pretty}")

        return True

    def _collect_child_keys_by_parent(self, dictionary):
        result = defaultdict(set)

        def walk(node, path):
            if isinstance(node, dict):
                if node:
                    result[path].update(node.keys())
                for k, v in node.items():
                    walk(v, f"{path}.{k}")
            elif isinstance(node, list):
                for item in node:
                    walk(item, path)

        walk(dictionary, self.ROOT_PATH)
        return result

    def _collect_values_by_path(self, dictionary):
        result = defaultdict(set)

        def is_primitive(x):
            return isinstance(x, (str, int, float, bool)) or x is None

        def walk(node, path):
            if is_primitive(node):
                result[path].add(node)
                return

            if isinstance(node, dict):
                for k, v in node.items():
                    walk(v, f"{path}.{k}")
                return

            if isinstance(node, list):
                for item in node:
                    walk(item, path)
                return

        walk(dictionary, self.ROOT_PATH)
        return result



from external_dependencies.caiman.mocked_motion_correction_submodule import MockedMotionCorrectionSubModule
from external_dependencies.caiman.mocked_movie import MockedMovie
from external_dependencies.caiman.mocked_source_extraction import MockedSourceExtractionModule


class InMemoryCaiman:
    def __init__(self, file_system=None):
        self._file_system = file_system
        self.motion_correction = MockedMotionCorrectionSubModule(file_system)
        self.source_extraction = MockedSourceExtractionModule(file_system)

    def load(self, fname):
        return MockedMovie(fname, self._file_system)


import io
import json

from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from ci_pipe.step import Step


class Plotter:
    def __init__(self, console=None):
        self.console = console or Console(file=io.StringIO(), force_terminal=True)

    def get_step_info(self, trace, step_number, branch, show_parameters=True):
        step = self._step_from(trace, step_number, branch)
        if not step:
            return
        table = self._build_table_from(step, step_number, show_parameters)
        self.console.print(table)

    def _step_from(self, trace, step_number, branch_name):
        try:
            steps = trace.steps_from(branch_name)
        except KeyError:
            self.console.print(f"[bold red]Branch '{branch_name}' not found in trace[/bold red]")
            return None

        for index, step in enumerate(steps, start=1):
            if index == step_number:
                return step

        self.console.print(f"[bold red]Step {step_number} not found in branch '{branch_name}'[/bold red]")
        return None

    def _build_table_from(self, step: Step, step_number, show_parameters):
        table = Table(title=f"Step {step_number} - {step.name()}", show_lines=True)
        table.add_column("Field", style="cyan", no_wrap=True)
        table.add_column("Value", style="yellow")

        params = step.arguments()
        outputs =step.output()

        table.add_row("Name", step.name())
        if show_parameters and params:
            params_str = "\n".join(f"{k}: {v}" for k, v in params.items())
            table.add_row("Parameters", params_str)

        outputs_str = json.dumps(outputs, indent=2)
        table.add_row("Outputs", outputs_str)

        return table

    def get_all_trace_from_branch(self, trace, branch_name):
        branch = trace.branch_from(branch_name)
        if not branch:
            self.console.print(f"[bold red]Branch '{branch_name}' not found[/bold red]")
            return

        steps = trace.steps_from(branch_name)
        if not steps:
            self.console.print(f"[yellow]No steps found in branch '{branch_name}'[/yellow]")
            return

        panels = self._build_trace_panels(steps)

        self.console.print(f"\n[bold underline]Pipeline Trace of branch: {branch_name}[/bold underline]\n")
        self.console.print(*panels, justify="center")

    def _format_params(self, params):
        if not params:
            return "None"
        return "\n".join(f"  {k}: {v}" for k, v in params.items())

    def _build_trace_panels(self, steps):
        panels = [
            Panel(
                f"[bold]{index}.[/bold] {step.name()}\n"
                f"Params:\n{self._format_params(step.arguments())}",
                padding=(1, 2),
            )
            for index, step in enumerate(steps, start=1)
        ]

        items = []
        for i, p in enumerate(panels):
            items.append(p)
            if i < len(panels) - 1:
                items.append("⬇")
        return items



class MultiModuleProxy:
    """
    A transparent proxy that intercepts method calls and broadcasts them 
    to the specific module of every pipeline in the MultiCIPipe.
    """
    def __init__(self, multi_pipe, module_name):
        self._multi_pipe = multi_pipe
        self._module_name = module_name

    def __getattr__(self, name):
        def dynamic_dispatch(*args, **kwargs):
            
            def action(pipeline):
                module = getattr(pipeline, self._module_name)
                
                if not hasattr(module, name):
                     raise AttributeError(f"Module '{self._module_name}' has no attribute '{name}'")
                
                method_to_call = getattr(module, name)
                method_to_call(*args, **kwargs)

            self._multi_pipe.with_pipelines_do(action)
            return self._multi_pipe

        return dynamic_dispatch


class MockedMovie:
    def __init__(self, fname, file_system):
        self._fname = fname
        self._file_system = file_system

    def save(self, output_path):
        self._file_system.write(output_path, "")



import yaml


class ConfigDefaults:
    """
    Responsible for loading default configuration values from a file.
    Currently supports YAML.
    """

    @staticmethod
    def load_from_file(path, file_system):
        """
        Loads defaults from a configuration file and returns them as a dictionary.

        Args:
            path (str): Path to the configuration file.

        Returns:
            dict: Parsed configuration data.

        Raises:
            FileNotFoundError: If the file does not exist.
            ValueError: If the file cannot be parsed or does not contain a dictionary.
        """
        if not file_system.exists(path):
            raise FileNotFoundError(f"Config file not found: {path}")

        try:
            content = file_system.read(path)
            data = yaml.safe_load(content) or {}
            if not isinstance(data, dict):
                raise ValueError(f"Config file {path} must contain a dictionary at the top level.")
            return data
        except yaml.YAMLError as e:
            raise ValueError(f"Error parsing YAML config: {e}")



class MockedCNMFModule:
    def __init__(self, file_system):
        self._file_system = file_system

    def CNMF(
            self,
            n_processes=None,
            k=5,
            gSig=[4, 4],
            gSiz=None,
            merge_thresh=0.8,
            p=2,
            dview=None,
            Ain=None,
            Cin=None,
            b_in=None,
            f_in=None,
            do_merge=True,
            ssub=2,
            tsub=2,
            p_ssub=1,
            p_tsub=1,
            method_init='greedy_roi',
            alpha_snmf=0.5,
            rf=None,
            stride=None,
            memory_fact=1,
            gnb=1,
            nb_patch=1,
            only_init_patch=False,
            method_deconvolution='oasis',
            n_pixels_per_process=4000,
            block_size_temp=5000,
            num_blocks_per_run_temp=20,
            num_blocks_per_run_spat=20,
            check_nan=True,
            skip_refinement=False,
            normalize_init=True,
            options_local_NMF=None,
            minibatch_shape=100,
            minibatch_suff_stat=3,
            update_num_comps=True,
            rval_thr=0.9,
            thresh_fitness_delta=-20,
            thresh_fitness_raw=None,
            thresh_overlap=.5,
            batch_update_suff_stat=False,
            s_min=None,
            remove_very_bad_comps=False,
            border_pix=0,
            low_rank_background=True,
            update_background_components=True,
            rolling_sum=True,
            rolling_length=100,
            min_corr=.85,
            min_pnr=20,
            ring_size_factor=1.5,
            center_psf=False,
            use_dense=True,
            deconv_flag=True,
            simultaneously=False,
            n_refit=0,
            del_duplicates=False,
            N_samples_exceptionality=None,
            max_num_added=3,
            min_num_trial=2,
            thresh_CNN_noisy=0.5,
            fr=30,
            decay_time=0.4,
            min_SNR=2.5,
            ssub_B=2,
            init_iter=2,
            sniper_mode=False,
            use_peak_max=False,
            test_both=False,
            expected_comps=500,
            params=None
    ):
        return MockedCNMFModule(
            file_system=self._file_system
        )

    def fit(self, images=None):
        pass

    def estimates(self):
        pass

    def save(self, path):
        self._file_system.write(path, "")



import os
import shutil
from typing import List

from .file_system_interface import FileSystemInterface

class PersistentFileSystem(FileSystemInterface):
    def write(self, path: str, content: str):
        with open(path, 'w', encoding='utf-8') as f:
            f.write(content)

    def read(self, path: str) -> str:
        with open(path, 'r', encoding='utf-8') as f:
            return f.read()

    def exists(self, path: str) -> bool:
        return os.path.exists(path)

    def makedirs(self, path: str, exist_ok: bool = False):
        os.makedirs(path, exist_ok=exist_ok)

    def listdir(self, path: str) -> List[str]:
        return [os.path.join(path, name) for name in os.listdir(path)]
    
    def subdirs(self, path: str) -> List[str]:
        return [name for name in os.listdir(path) if os.path.isdir(os.path.join(path, name))]

    def open(self, path: str, mode: str = 'r', encoding: str = None):
        return open(path, mode, encoding=encoding)

    def copy2(self, src: str, dst: str):
        return shutil.copy2(src, dst)

    def join(self, directory: str, filename: str):
        return os.path.join(directory, filename)

    def base_path(self, path):
        return os.path.basename(path)

    def split_text(self, path):
        return os.path.splitext(path)
    
    def remove(self, path):
        return os.remove(path)


from io import StringIO
from typing import List

from .file_system_interface import FileSystemInterface

class InMemoryFileSystem(FileSystemInterface):
    def __init__(self):
        self.files = {}
        self.directories = set()

    def write(self, path: str, content: str):
        from io import StringIO
        self.files[path] = StringIO(content)

    def read(self, path: str) -> str:
        file_obj = self.files.get(path, None)
        if file_obj is None:
            raise FileNotFoundError(f"No such file: {path}")
        file_obj.seek(0)
        return file_obj.read()

    def exists(self, path: str) -> bool:
        return path in self.files or path in self.directories

    def makedirs(self, path: str, exist_ok: bool = False):
        self.directories.add(path)

    def listdir(self, path: str) -> List[str]:
        if path not in self.directories:
            raise FileNotFoundError(f"No such directory: {path}")
        return [f for f in self.files if f.startswith(path)]
    
    def subdirs(self, path: str) -> List[str]:
        if path not in self.directories:
            raise FileNotFoundError(f"No such directory: {path}")
        subdirs = set()
        prefix = path if path.endswith("/") else path + "/"
        for file_path in self.files:
            if file_path.startswith(prefix):
                remainder = file_path[len(prefix):]
                subdir = remainder.split("/", 1)[0]
                subdirs.add(subdir)
        return list(subdirs)

    def open(self, path: str, mode: str = 'r', encoding: str = None):
        if 'w' in mode:
            file_content = ""
            self.files[path] = StringIO(file_content)
        elif 'r' in mode:
            content = self.files.get(path, None)
            if content is None:
                raise FileNotFoundError(f"No such file: {path}")
            return content
        return None

    def copy2(self, src: str, dst: str):
        if src in self.files:
            src_filename = self.base_path(src)
            dst_path = self.join(dst, src_filename)
            self.files[dst_path] = StringIO(self.files[src].getvalue())
            return dst_path
        else:
            raise FileNotFoundError(f"No such file: {src}")

    def join(self, directory, filename):
        return f"{directory}/{filename}"

    def base_path(self, path):
        return path.split("/")[-1]

    def split_text(self, path):
        if "." in path:
            idx = path.rfind(".")
            return (path[:idx], path[idx+1:])
        return (path, "")
    
    def remove(self, path):
        if path in self.files:
            del self.files[path]
        else:
            raise FileNotFoundError(f"No such file: {path}")